;################################################################################
;#										#
;# libmio - multi i/o for ATMega 16/32 	version 0.24				#
;# copyright (c) 2005/2006/2007 Joerg Wolfram (joerg@jcwolfram.de)		#
;#										#
;# * PAL/NTSC RGB videodisplay with 29x23 chars and 8 colors			#
;# * 1 channel sound can play notes and noise					#
;# * ps2-keyboard interface							#
;# * 1200 Bps bidirectional serial port						#
;# * parallel printer port							#
;# * pseudorandom generator							#
;# * 0,1s timer									#
;#										#
;# This library is free software; you can redistribute it and/or		#
;# modify it under the terms of the GNU Lesser General Public			#
;# License as published by the Free Software Foundation; either			#
;# version 2 of the License, or (at your option) any later version.		#
;#										#
;# This library is distributed in the hope that it will be useful,		#
;# but WITHOUT ANY WARRANTY; without even the implied warranty of		#
;# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.	See the GNU		#
;# Lesser General Public License for more details.				#
;#										#
;# You should have received a copy of the GNU Lesser General Public		#
;# License along with this library; if not, write to the			#
;# Free Software Foundation, Inc., 59 Temple Place - Suite 330,			#
;# Boston, MA 02111-1307, USA.							#
;#										#
;################################################################################

;-------------------------------------------------------------------------------
;macro for output a color character cell (30 cycles)
;-------------------------------------------------------------------------------
.macro libmio_oc8
		out	PORTC,tempreg1		;1 out P0
		ld	ZL,X+			;2 get char
		sbrc	tempreg2,6		;1 check P1 
		swap	tempreg1		;1
		
		out	PORTC,tempreg1		;1 out P1
		mov	r0,tempreg1		;2 get color for P5
		nop				;1 
		sbrc	tempreg2,5		;1 check P2
		swap	tempreg1		;1
		
		out	PORTC,tempreg1		;1 out P2
		ldd	tempreg3,Y+2		;2 get color++
		sbrc	tempreg2,3		;1 check P3
		swap	tempreg1		;1

		out	PORTC,tempreg1		;1 out P3
		sbrc	tempreg2,2		;1 check P4
		swap	tempreg1		;1
		sbrc	tempreg2,1		;1 check P5
		swap	r0			;1
		
		out	PORTC,tempreg1		;1 out P4
		lpm	tempreg2,Z		;3 get pixline
		nop				;1
		
		out	PORTC,r0		;1 out P5
		ld	tempreg1,Y+		;2 get color
		sbrc	tempreg2,7		;1 check P4
		swap	tempreg1		;1
.endm

;-------------------------------------------------------------------------------
;macro for output a mono character cell (24 cycles)
;-------------------------------------------------------------------------------
.macro libmio_oc2a
		out	PORTC,tempreg1		;1 out P0
		sbrc	tempreg2,6		;1 check P1 
		swap	tempreg1		;1
		mov	r0,tempreg1		;1
		
		out	PORTC,tempreg1		;1 out P1
		sbrc	tempreg2,5		;1 check P2
		swap	r0			;1
		mov	tempreg1,YL		;1 color
		
		out	PORTC,r0		;1 out P2
		sbrc	tempreg2,3		;1 check P3
		swap	r0			;1

		out	PORTC,r0		;1 out P3
		sbrc	tempreg2,2		;1 check P4
		swap	r0			;1
		nop				;1
		
		out	PORTC,r0		;1 out P4
		ld	ZL,X+			;2
		
		out	PORTC,tempreg1		;1 out P5=0
		lpm	tempreg2,Z		;3 get char
		sbrc	tempreg2,7		;1 check P4
		swap	tempreg1		;1
.endm

;-------------------------------------------------------------------------------
;macro for output a mono character cell (24 cycles)
;-------------------------------------------------------------------------------
.macro libmio_oc2b
		out	PORTC,tempreg1		;1 out P0
		sbrc	tempreg2,6		;1 check P1 
		swap	tempreg1		;1
		mov	r0,tempreg1		;1
		
		out	PORTC,tempreg1		;1 out P1
		sbrc	tempreg2,5		;1 check P2
		swap	r0			;1
		mov	tempreg1,r12		;1 color
		
		out	PORTC,r0		;1 out P2
		sbrc	tempreg2,3		;1 check P3
		swap	r0			;1

		out	PORTC,r0		;1 out P3
		sbrc	tempreg2,2		;1 check P4
		swap	r0			;1
		nop				;1
		
		out	PORTC,r0		;1 out P4
		ld	ZL,X+			;2
		
		out	PORTC,tempreg1		;1 out P5=0
		lpm	tempreg2,Z		;3 get char
		sbrc	tempreg2,7		;1 check P4
		swap	tempreg1		;1
.endm
;-------------------------------------------------------------------------------
;macro for output a mono character cell (24 cycles)
;-------------------------------------------------------------------------------
.macro libmio_oc2c
		out	PORTC,tempreg1		;1 out P0
		sbrc	tempreg2,6		;1 check P1 
		swap	tempreg1		;1
		mov	r0,tempreg1		;1
		
		out	PORTC,tempreg1		;1 out P1
		sbrc	tempreg2,5		;1 check P2
		swap	r0			;1
		mov	tempreg1,r13		;1 color
		
		out	PORTC,r0		;1 out P2
		sbrc	tempreg2,3		;1 check P3
		swap	r0			;1

		out	PORTC,r0		;1 out P3
		sbrc	tempreg2,2		;1 check P4
		swap	r0			;1
		nop				;1
		
		out	PORTC,r0		;1 out P4
		ld	ZL,X+			;2
		
		out	PORTC,tempreg1		;1 out P5=0
		lpm	tempreg2,Z		;3 get char
		sbrc	tempreg2,7		;1 check P4
		swap	tempreg1		;1
.endm
;-------------------------------------------------------------------------------
;macro for output a mono character cell (24 cycles)
;-------------------------------------------------------------------------------
.macro libmio_oc2d
		out	PORTC,tempreg1		;1 out P0
		sbrc	tempreg2,6		;1 check P1 
		swap	tempreg1		;1
		mov	r0,tempreg1		;1
		
		out	PORTC,tempreg1		;1 out P1
		sbrc	tempreg2,5		;1 check P2
		swap	r0			;1
		mov	tempreg1,r14		;1 color
		
		out	PORTC,r0		;1 out P2
		sbrc	tempreg2,3		;1 check P3
		swap	r0			;1

		out	PORTC,r0		;1 out P3
		sbrc	tempreg2,2		;1 check P4
		swap	r0			;1
		nop				;1
		
		out	PORTC,r0		;1 out P4
		ld	ZL,X+			;2
		
		out	PORTC,tempreg1		;1 out P5=0
		lpm	tempreg2,Z		;3 get char
		sbrc	tempreg2,7		;1 check P4
		swap	tempreg1		;1
.endm
;-------------------------------------------------------------------------------
;macro for output a mono character cell (24 cycles)
;-------------------------------------------------------------------------------
.macro libmio_oc2e
		out	PORTC,tempreg1		;1 out P0
		sbrc	tempreg2,6		;1 check P1 
		swap	tempreg1		;1
		mov	r0,tempreg1		;1
		
		out	PORTC,tempreg1		;1 out P1
		sbrc	tempreg2,5		;1 check P2
		swap	r0			;1
		mov	tempreg1,r15		;1 color
		
		out	PORTC,r0		;1 out P2
		sbrc	tempreg2,3		;1 check P3
		swap	r0			;1

		out	PORTC,r0		;1 out P3
		sbrc	tempreg2,2		;1 check P4
		swap	r0			;1
		nop				;1
		
		out	PORTC,r0		;1 out P4
		ld	ZL,X+			;2
		
		out	PORTC,tempreg1		;1 out P5=0
		lpm	tempreg2,Z		;3 get char
		sbrc	tempreg2,7		;1 check P4
		swap	tempreg1		;1
.endm

;-------------------------------------------------------------------------------
; -output character line 5 clocks per pixel
; -256 characters are in table 6x10px each
;-------------------------------------------------------------------------------
libmio_tvc:	lds	XL,libmio_rampos_l	;2 BS low
		lds	XH,libmio_rampos_h	;1 BS high
		movw	YL,XL			;2 copy values
		ldi	ZL,LOW(690)		;1 offset color table
		ldi	ZH,HIGH(690)		;1
		add	YL,ZL			;1
		adc	YH,ZH			;1
	
		ldi	ZH,HIGH(libmio_ctable*2)	;start chartable
		lds	ZL,libmio_cline		;2 character line
		add	ZH,ZL			;1

	    	ld	ZL,X+			;2 get char
		ld	tempreg1,Y+		;2 get color	
		lpm	tempreg2,Z		;3 get pixline
		ld	tempreg3,Y		;2 next color
		bst	tempreg3,0		;1 fsel bit
		sbrc	tempreg2,7		;1 check P4
		swap	tempreg1		;1

		libmio_oc8			;char 1
		libmio_oc8			;char 2
		libmio_oc8			;char 3
		libmio_oc8			;char 4
		libmio_oc8			;char 5
		libmio_oc8			;char 6
		libmio_oc8			;char 7
		libmio_oc8			;char 8
		libmio_oc8			;char 9
		libmio_oc8			;char 10
		libmio_oc8			;char 11
		libmio_oc8			;char 12
		libmio_oc8			;char 13
		libmio_oc8			;char 14
		libmio_oc8			;char 15
		libmio_oc8			;char 16
		libmio_oc8			;char 17
		libmio_oc8			;char 18
		libmio_oc8			;char 19
		libmio_oc8			;char 20
		libmio_oc8			;char 21
		libmio_oc8			;char 22
		libmio_oc8			;char 23
		libmio_oc8			;char 24
		libmio_oc8			;char 25
		libmio_oc8			;char 26
		libmio_oc8			;char 27
		libmio_oc8			;char 28
		libmio_oc8			;char 29
		libmio_oc8			;char 30
		out	PORTC,byte0		;1 set to border
		ld	ZL,-X			;2 correct X-register
		lds	ZH,libmio_cline		;2 character line
		inc	ZH
		cpi	ZH,10			;1 check
		brne	libmio_tvc_1		;2 jump if not next line
		clr	ZH			;1 character line = 0
		sts	libmio_rampos_l,XL	;2 new memory position
		sts	libmio_rampos_h,XH	;2 new memory position
libmio_tvc_1:	sts	libmio_cline,ZH		;2 new character line
		jmp	libmio_is_15


;-------------------------------------------------------------------------------
; -output character line 4 clocks per pixel
; -256 characters are in table 6x12px each
;-------------------------------------------------------------------------------
libmio_tvm:	lds	XL,libmio_rampos_l	;2 BS low
		lds	XH,libmio_rampos_h	;1 BS high
		lds	YL,libmio_elcolor	;2 recent line color
		mov	YH,vline_l		;1 get pixel line
		ldi	ZH,HIGH(libmio_ectable*2)	;start chartable
		lds	ZL,libmio_cline		;2 character line
		add	ZH,ZL			;1
		inc	ZL			;1 cline+1
		sts	libmio_cline,ZL		;2 store back character line
		cpi	ZL,10
		brcs	libmio_tvm5		;a visible line
		cpi	ZL,12
		brne	libmio_tvm4
		adiw	XL,35			;1 next line
		sts	libmio_rampos_l,XL	;2 new memory position
		sts	libmio_rampos_h,XH	;2 new memory position
		sts	libmio_cline,byte0	;2 clear cline



libmio_tvm1:	cpi	YH,12			;1 start of second line
		brne	libmio_tvm2		;no		
		lds	YL,libmio_estatus	;get color of status line
		sts	libmio_elcolor,YL
		jmp	libmio_is_15



libmio_tvm2:	cpi	YH,24			;1 last of head
		brne	libmio_tvm3
		ldi	XL,LOW(libmio_vram+3*libmio_ecols)
		ldi	XH,HIGH(libmio_vram+3*libmio_ecols)
		lds	YL,libmio_edstart
		ldi	YH,libmio_ecols
		mul	YL,YH
		add	XL,r0
		adc	XH,r1
		ldi	YL,0x0e			;1 white on black
		sts	libmio_rampos_l,XL	;2 new memory position
		sts	libmio_rampos_h,XH	;2 new memory position
		sts	libmio_elcolor,YL
		jmp	libmio_is_15		


libmio_tvm3:	cpi	YH,216			;start of help line
		brne	libmio_tvm4
		ldi	YL,0xe4			;white on red
		ldi	XL,LOW(libmio_vram+2*libmio_ecols)
		ldi	XH,HIGH(libmio_vram+2*libmio_ecols)
		sts	libmio_rampos_l,XL	;2 new memory position
		sts	libmio_rampos_h,XH	;2 new memory position
		sts	libmio_elcolor,YL
libmio_tvm4:	jmp	libmio_is_15		

libmio_tvm5:	push	r12
		push	r13
		push	r14
		push	r15
		cpi	YH,216
		brcc	libmio_tvm6
		mov	r12,YL			;copy color
		mov	r13,YL			;copy color
		mov	r14,YL			;copy color
		mov	r15,YL			;copy color
		nop
		nop
		nop
		nop
		nop
		rjmp	libmio_tvm7

libmio_tvm6:	ldi	YL,0x6e			;white on magenta
		mov	r15,YL
		ldi	YL,0x2e			;white on blue
		mov	r14,YL
		ldi	YL,0x80			;black on green
		mov	r13,YL
		ldi	YL,0xc0			;black on yellow
		mov	r12,YL
		ldi	YL,0x4e			;white on red
		

libmio_tvm7:	ld	ZL,X+			;2 get char
		lpm	tempreg2,Z		;3 get pixline
		mov	tempreg1,YL		;2 get color	
		sbrc	tempreg2,7		;1 check P4
		swap	tempreg1		;1

		libmio_oc2a			;char 2
		libmio_oc2a			;char 3
		libmio_oc2a			;char 4
		libmio_oc2a			;char 5
		libmio_oc2a			;char 6
		libmio_oc2a			;char 7
		libmio_oc2b			;char 8
		libmio_oc2b			;char 9
		libmio_oc2b			;char 10
		libmio_oc2b			;char 11
		libmio_oc2b			;char 12
		libmio_oc2b			;char 13
		libmio_oc2b			;char 14
		libmio_oc2c			;char 15
		libmio_oc2c			;char 16
		libmio_oc2c			;char 17
		libmio_oc2c			;char 18
		libmio_oc2c			;char 19
		libmio_oc2c			;char 20
		libmio_oc2c			;char 21
		libmio_oc2d			;char 22
		libmio_oc2d			;char 23
		libmio_oc2d			;char 24
		libmio_oc2d			;char 25
		libmio_oc2d			;char 26
		libmio_oc2d			;char 27
		libmio_oc2d			;char 28
		libmio_oc2e			;char 29
		libmio_oc2e			;char 30
		libmio_oc2e			;char 31
		libmio_oc2e			;char 32
		libmio_oc2e			;char 33
		libmio_oc2e			;char 34
		libmio_oc2e			;char 35
		libmio_oc2e			;char 35
		out	PORTC,byte0		;1 set to border
		pop	r15
		pop	r14
		pop	r13
		pop	r12
		jmp	libmio_is_15

;-------------------------------------------------------------------------------
; -graphics mode 168x116 2 colors
;-------------------------------------------------------------------------------
libmio_tvg2:	lds	XL,libmio_rampos_l	;2 BS low
		lds	XH,libmio_rampos_h	;1 BS high

		lds	YH,0x0100		;get color 0
		lds	YL,0x0101		;get color 1
		swap	YL			;swap color 1
		add	YH,YL
		andi	YH,0xee
		ldi	ZL,21			;number of bytes
		
		ld	ZH,X+			;get first byte
		mov	tempreg1,YH		;background
		sbrc	ZH,0			;check bit 1
		swap	tempreg1		;set to foreground

libmio_tvg2_1:	out	PORTC,tempreg1		;1 out bit 1		
		nop				;1 fill		
		mov	tempreg1,YH		;1 background
		sbrc	ZH,1			;1 check bit 2
		swap	tempreg1		;1 set to foreground
	
		out	PORTC,tempreg1		;1 out bit 2		
		nop				;1 fill		
		mov	tempreg1,YH		;1 background
		sbrc	ZH,2			;1 check bit 3
		swap	tempreg1		;1 set to foreground

		out	PORTC,tempreg1		;1 out bit 3		
		nop				;1 fill		
		mov	tempreg1,YH		;1 background
		sbrc	ZH,3			;1 check bit 4
		swap	tempreg1		;1 set to foreground
		nop				;1 fill

		out	PORTC,tempreg1		;1 out bit 4		
		mov	tempreg3,YH		;1 background		
		mov	tempreg1,YH		;1 background
		sbrc	ZH,4			;1 check bit 5
		swap	tempreg1		;1 set to foreground
		
		out	PORTC,tempreg1		;1 out bit 5		
		ld	YL,X+			;2 get next byte
		sbrc	ZH,5			;1 check bit 6
		swap	tempreg3		;1 set to foreground
		mov	tempreg2,YH		;1 background
		
		out	PORTC,tempreg3		;1 out bit 6
		sbrc	ZH,6			;1 check bit 7
		swap	tempreg2		;1 set to foreground
		mov	tempreg3,YH		;1 background		
		mov	tempreg1,YH		;1 background

		out	PORTC,tempreg2		;1 out bit 7
		sbrc	ZH,7			;1 check bit 8
		swap	tempreg3		;1 set to foreground
		sbrc	YL,0			;1 check bit 1 of next byte
		swap	tempreg1		;1 set to foreground
		nop				;1 fill
		
		out	PORTC,tempreg3		;1 out bit 8
		mov	ZH,YL			;1 copy byte
		dec	ZL			;2 loop counter
		brne 	libmio_tvg2_1	
		
		out	PORTC,byte0		;1 set to border
		
		ld	ZL,-X			;2 correct X-register
		lds	ZH,libmio_cline		;2 character line
		inc	ZH
		cpi	ZH,2			;1 check
		brne	libmio_tvg2_2		;2 jump if not next line
		clr	ZH			;1 character line = 0
		sts	libmio_rampos_l,XL	;2 new memory position
		sts	libmio_rampos_h,XH	;2 new memory position
libmio_tvg2_2:	sts	libmio_cline,ZH		;2 new character line
		jmp	libmio_is_15


;-------------------------------------------------------------------------------
; -graphics mode 120x76 4 colors
;-------------------------------------------------------------------------------
libmio_tvg4:	lds	XL,libmio_rampos_l	;2 BS low
		lds	XH,libmio_rampos_h	;1 BS high

		ldi	YH,0x01			;1 start of palette
		ldi	ZH,30			;1 number of bytes
				
		ld	ZL,X+			;2 get first byte
		mov	YL,ZL			;1 copy
		andi	YL,0x03			;1
		ld	tempreg2,Y		;2 get pixel 1
		
libmio_tvg4_1:	out	PORTC,tempreg2		;1 output pixel 1
		mov	YL,ZL			;1 copy back 
		lsr	YL			;1
		lsr	YL			;1
		andi	YL,0x03			;1
		ld	tempreg2,Y		;2 get pixel 2
		swap	ZL			;1
		
		out	PORTC,tempreg2		;1 output pixel 2
		mov	YL,ZL			;1 copy back 
		andi	YL,0x03			;1
		ld	tempreg2,Y		;2 get pixel 3
		mov	YL,ZL			;1
		lsr	YL			;1
		
		out	PORTC,tempreg2		;1 output pixel 3
		lsr	YL			;1
		andi	YL,0x03			;1
		ld	tempreg2,Y		;2 get pixel 4
		ld	ZL,X+			;2 get next byte
		mov	YL,ZL			;1 copy
		
		out	PORTC,tempreg2		;1 output pixel 4
		andi	YL,0x03			;1
		ld	tempreg2,Y		;2 get pixel 1
		dec	ZH			;1 loop counter
		brne 	libmio_tvg4_1		;2
		
		out	PORTC,byte0		;1 set to border
		
		ld	ZL,-X			;2 correct X-register
		lds	ZH,libmio_cline		;2 character line
		inc	ZH
		cpi	ZH,3			;1 check
		brne	libmio_tvg4_2		;2 jump if not next line
		clr	ZH			;1 character line = 0
		sts	libmio_rampos_l,XL	;2 new memory position
		sts	libmio_rampos_h,XH	;2 new memory position
libmio_tvg4_2:	sts	libmio_cline,ZH		;2 new character line
		jmp	libmio_is_15


;-------------------------------------------------------------------------------
; -graphics mode 84x58 8 colors
;-------------------------------------------------------------------------------
libmio_tvg8:	lds	XL,libmio_rampos_l	;2 BS low
		lds	XH,libmio_rampos_h	;1 BS high
		push	r16			;2 save reg

		ldi	ZL,42			;number of bytes
		ldi	YH,0x01			;start of palette
		ld	YL,X+			;get first byte
		lsr	YL			;1
		mov	r16,YL			;1
		swap	YL			;lower nibble first
		andi	YL,0x07			;mask bits
		ld	ZH,Y			;get value
		
libmio_tvg8_1:	out	PORTC,ZH		;1 out pix 0
		mov	YL,r16			;1
		nop				;1
		andi	YL,0x07			;1 mask bits
		ld	ZH,Y			;2 get value
		nop				;1 fill	
		nop				;1 fill	
		nop				;1 fill		
		out	PORTC,ZH		;1 out pix 1		

		ld	YL,X+			;2 get next byte
		lsr	YL			;1
		mov	r16,YL			;1
		swap	YL			;1 lower nibble first
		andi	YL,0x07			;1 mask bits
		ld	ZH,Y			;2 get value		

		dec	ZL			;1 loop counter
		brne 	libmio_tvg8_1		;2 loop
		
		out	PORTC,byte0		;1 set to border

		sbiw	XL,1			;correct
		pop	r16			;restore
		lds	ZH,libmio_cline		;2 character line
		inc	ZH
		cpi	ZH,4			;1 check
		brne	libmio_tvg8_2		;2 jump if not next line
		clr	ZH			;1 character line = 0
		sts	libmio_rampos_l,XL	;2 new memory position
		sts	libmio_rampos_h,XH	;2 new memory position
libmio_tvg8_2:	sts	libmio_cline,ZH		;2 new character line
		jmp	libmio_is_15
