;################################################################################
;#										#
;# avr-chipbasic2 - single chip basic computer with ATmega644			#
;# dataflash file manager 							#
;# copyright (c) 2006/2007 Joerg Wolfram (joerg@jcwolfram.de)			#
;#										#
;#										#
;# This program is free software; you can redistribute it and/or		#
;# modify it under the terms of the GNU General Public License			#
;# as published by the Free Software Foundation; either version 2		#
;# of the License, or (at your option) any later version.			#
;#										#
;# This program is distributed in the hope that it will be useful,		#
;# but WITHOUT ANY WARRANTY; without even the implied warranty of		#
;# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.See the GNU		#
;# General Public License for more details.					#
;#										#
;# You should have received a copy of the GNU General Public			#
;# License along with this library; if not, write to the			#
;# Free Software Foundation, Inc., 59 Temple Place - Suite 330,			#
;# Boston, MA 02111-1307, USA.							#
;#										#
;################################################################################

;------------------------------------------------------------------------------
;main screen mask
;------------------------------------------------------------------------------
confp_main:	ldi	XL,1
		libmio_chpump			;enable chargepump for serial
		sts	libmio_channel,byte0	;output to screen
		sts	libmio_vidmode,byte0	;standard video mode
		ldi	XL,0x0e			;color scheme (white on black)
		sts	libmio_color,XL
		ldi	YL,0			;maximal sized box
		ldi	YH,0
		ldi	ZL,29
		ldi	ZH,22
		libmio_whbox
		ldi	XL,0xe0			;color scheme (white on black)
		sts	libmio_color,XL
		libmio_thistext
		.db 1,2,"AVR-ChipBasic2 configpage",0
confp_main_02:	ldi	XL,0x27			;white on red
		libmio_color
		libmio_thistext
		.db 22,0," Exit ",0,0		;show message
		ldi	XL,0x60			;black on yellow
		libmio_color
		libmio_thistext
		.db 22,6," Keyb ",0,0		;show message
		ldi	XL,0x40			;black on green
		libmio_color
		libmio_thistext
		.db 22,12,"      ",0,0		;show message
		ldi	XL,0x17			;white on blue
		libmio_color
		libmio_thistext
		.db 22,18,"      ",0,0		;show message
		ldi	XL,0x6e			;white on magenta
		sts	libmio_vram+690+689,XL
		ldi	XL,0x20			;space
		sts	libmio_vram+689,XL
		ldi	XL,0x37			;white on magenta
		libmio_color
		libmio_thistext
		.db 22,24,"Clone",0		;show message

 		ldi	XL,0x0f			;color scheme (white on black)
		sts	libmio_color,XL
		libmio_thistext
		.db 20,1,"Checksum: ",0,0
	    	rcall	confp_chksum	
		
;------------------------------------------------------------------------------
;the main loop
;------------------------------------------------------------------------------
confp_main_03:	rcall	confp_kbdtype		;flash filelist
confp_main_09:	libmio_waitkey


;------------------------------------------------------------------------------
;esc key
;------------------------------------------------------------------------------
confp_main_30:	cpi	tempreg1,0xed		;esc
		brne	confp_main_50
		rjmp	confp_wnk		;wait for no key and exit
		

;------------------------------------------------------------------------------
;F1 key (change keyboard layout)
;------------------------------------------------------------------------------
confp_main_50:	cpi	tempreg1,0xf1		;F1
		brne	confp_main_60
		lds	tempreg1,libmio_ksel	;recent layout
		inc	tempreg1
		cpi	tempreg1,0x03
		brne	confp_main_51
		clr	tempreg1
confp_main_51:	sts	libmio_ksel,tempreg1
		call	mem_setktype
		rcall	confp_kbdtype		;update list
		call	confp_wnk
		rjmp	confp_main_09

;------------------------------------------------------------------------------
;F2 key
;------------------------------------------------------------------------------
confp_main_60:	cpi	tempreg1,0xf2		;F2
		brne	confp_main_70

		rjmp	confp_main_09

;------------------------------------------------------------------------------
;F3 key
;------------------------------------------------------------------------------
confp_main_70:	cpi	tempreg1,0xf3		;F3
		brne	confp_main_80
		
		rjmp	confp_main_09

;------------------------------------------------------------------------------
;F4 key
;------------------------------------------------------------------------------
confp_main_80:	cpi	tempreg1,0xf4		;F4
		brne	confp_main_90
		call	clone_main
confp_main_90:	rjmp	confp_main


;------------------------------------------------------------------------------
;show keyboard layout
;------------------------------------------------------------------------------
confp_kbdtype: 	ldi	XL,0x0c			;color scheme (yellow on black)
		sts	libmio_color,XL
		libmio_thistext
		.db 3,1,"Keyboard layout:",0,0
	    	ldi	XL,0x17			;white on blue
		lds	tempreg1,libmio_ksel	;get active layout
		cpi	tempreg1,0x00
		brne	confp_kbdtype1
		ldi	XL,0x71			;blue on white
confp_kbdtype1:	libmio_color
	 	libmio_thistext
		.db 3,18,"DE (QWERTZ)",0

	    	ldi	XL,0x17			;white on blue
		lds	tempreg1,libmio_ksel	;get active layout
		cpi	tempreg1,0x01
		brne	confp_kbdtype2
		ldi	XL,0x71			;blue on white
confp_kbdtype2:	libmio_color
	 	libmio_thistext
		.db 4,18,"US (QWERTY)",0
		
	    	ldi	XL,0x17			;white on blue
		lds	tempreg1,libmio_ksel	;get active layout
		cpi	tempreg1,0x02
		brne	confp_kbdtype3
		ldi	XL,0x71			;blue on white
confp_kbdtype3:	libmio_color
	 	libmio_thistext
		.db 5,18,"DE (QWERTZ)",0
		ldi	XL,0x0e			;color scheme (white on black)
		sts	libmio_color,XL
	    	ret

;------------------------------------------------------------------------------
;wait for key release
;------------------------------------------------------------------------------
confp_wnk:	lds	XL,libmio_keycode
		cpi	XL,0
		brne	confp_wnk		;wait for key release
		ret

;------------------------------------------------------------------------------
;calc and output checksum
;------------------------------------------------------------------------------
confp_chksum:	clr	XL			;sum
		clr	XH
		clr	YL
		clr	YH
		clr	ZL			;ptr
		clr	ZH			
		
confp_chksum1:	lpm	r1,Z+
		ldi	tempreg1,2
		mul	r1,tempreg1	
		adc	XH,r0
		adc	YL,r1
		adc	YH,byte0
		lpm	tempreg1,Z+
		ldi	tempreg1,7
		mul	r1,tempreg1	
		adc	XH,r0
		adc	YL,r1
		adc	YH,byte0
		cpi	ZH,0x48
		brne	confp_chksum1
		
		clr	ZL			;ptr
		ldi	ZH,$f0			;		
		
confp_chksum2:	lpm	tempreg1,Z+
		ldi	tempreg1,113
		mul	r1,tempreg1	
		adc	XH,r0
		adc	YL,r1
		adc	YH,byte0
		lpm	tempreg1,Z+
		ldi	tempreg1,7
		mul	r1,tempreg1	
		adc	XH,r0
		adc	YL,r1
		adc	YH,byte0
		cpi	ZH,0x00
		brne	confp_chksum2
		push	XH
		push	XL
		movw	XL,YL
		ldi	ctrl,0x0f
		libmio_outhex
		pop	XL
		pop	XH
		libmio_outhex
		ret	
				